/*
 * interdiff - create incremental patch between two against a common source
 * Copyright (C) 2000, 2001 Tim Waugh <twaugh@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_ERROR_H
# include <error.h>
#endif /* HAVE_ERROR_H */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif /* HAVE_UNISTD_H */
#include <getopt.h>
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif /* HAVE_SYS_TYPES_H */
#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif /* HAVE_SYS_WAIT_H */

#include "util.h"
#include "diff.h"

#ifndef DIFF
#define DIFF "diff"
#endif

#ifndef PATCH
#define PATCH "patch"
#endif

struct file_list 
{
	char *file;
	long pos;
	struct file_list *next;
};

struct lines {
	char *line;
	unsigned long n;
	struct lines *next;
	struct lines *prev;
};

struct lines_info {
	struct lines *head;
	struct lines *tail;
};

static int human_readable = 1;
static int ignore_whitespace = 0;
static int max_context_real = 2, max_context = 2;
static int ignore_components = 0;
static int unzip = 0;

static struct patlist *pat_drop_context = NULL;

static struct file_list *files_done = NULL;
static struct file_list *files_in_patch2 = NULL;

/* checks whether file needs processed and sets context */
static int check_filename (const char *fn)
{
	if (patlist_match(pat_drop_context, fn)) {
		max_context = 0;
	} else {
		max_context = max_context_real;
	}
	return 1;
}

static void add_to_list (struct file_list **list, const char *file, long pos)
{
	struct file_list *make;
	make = xmalloc (sizeof *make);
	make->next = *list;
	make->file = xstrdup (file);
	make->pos = pos;
	*list = make;
}

static const char *stripped (const char *name)
{
	const char *basename;
	int i = 0;
	if (!strncmp (name, "./", 2))
		name += 2;
	basename = strrchr (name, '/');
	if (!basename)
		basename = name;
	while (i < ignore_components &&
	       (name = strchr (name, '/')) != NULL) {
		while (*name == '/')
			name++;
		i++;
	}
	return name ? name : basename;
}

static long file_in_list (struct file_list *list, const char *file)
{
	file = stripped (file);
	while (list) {
		if (!strcmp (stripped (list->file), file))
			return list->pos;
		list = list->next;
	}
	return -1;
}

static void free_list (struct file_list *list)
{
	struct file_list *next;
	for (; list; list = next) {
		next = list->next;
		free (list->file);
		free (list);
	}
}

static void insert_line_before (struct lines_info *lines,
				struct lines *at, struct lines *make)
{
	make->next = at;
	make->prev = at->prev;
	if (at->prev)
		at->prev->next = make;
	else
		lines->head = make;
	at->prev = make;
}

static int add_line (struct lines_info *lines, const char *line,
		     unsigned long n)
{
	struct lines *make;
	struct lines *at;
	
	make = xmalloc (sizeof *make);
	make->n = n;
	make->line = xstrdup (line);

	if (!lines->tail) {
		/* New */
		make->prev = NULL;
		make->next = NULL;
		lines->head = make;
		lines->tail = make;
		return 0;
	}
	if (lines->tail->n < n) {
		/* End */
		lines->tail->next = make;
		make->next = NULL;
		make->prev = lines->tail;
		lines->tail = make;
		return 0;
	}

	if (!lines->head)
		error (EXIT_FAILURE, 0, "List corrupted: no head");

	at = lines->head;
	while (at) {
		if (n < at->n) {
			insert_line_before (lines, at, make);
			return 0;
		}
		if (n == at->n) {
			free (make->line);
			free (make);
			if (!strcmp (at->line, line))
				/* Already there. */
				return 0;

			/* Clash: first one wins. */
			return 1;
		}
		make->prev = at;
		at = at->next;
	}

	if (!at)
		error (EXIT_FAILURE, 0, "List corrupted: ordering");

	return 1;
}

static void merge_lines (struct lines_info *lines1, struct lines_info *lines2)
{
	struct lines *at, *at_prev, *make;


	/* 
	 * Note: we don't care about the tail member here - not needed
	 * anymore.
	 */

	if (!lines1->head) {
		/* first list empty - only take second */
		lines1->head = lines2->head;
		return;
	}

	/* merge lines in one pass */
	at = lines1->head;
	make = lines2->head;
	while (at) {
		while (make) {
			if (make->n > at->n)
				break;

			if (make->n < at->n) {
				struct lines *next = make->next;

				insert_line_before (lines1, at, make);
				make = next;
				continue;
			}

			make = make->next; /* line number equal - first wins */
		}
		at_prev = at;
		at = at->next;
	}

	/* append the rest of list 2 to list 1 */
	if (make)
		at_prev->next = make;
}

static void free_lines (struct lines *list)
{
	struct lines *next;
	for (; list; list = next) {
		next = list->next;
		free (list->line);
		free (list);
	}
}

static struct lines *create_orig (FILE *f, struct lines_info *file)
{
	unsigned long linenum;
	char *line = NULL;
	ssize_t linelen = 0;
	int last_was_add;
	long pos = ftell (f);

	do {
		if (getline (&line, &linelen, f) == -1)
			break;
	} while (strncmp (line, "@@ ", 3));

	while (!feof (f)) {
		/* Find next hunk */
		unsigned long orig_lines, new_lines, newline;
		int file_is_removed = 0;
		char *p, *q;

		if (strncmp (line, "@@", 2)) {
			fseek (f, pos, SEEK_SET);
			break;
		}

		p = strchr (line, '-');
		if (!p)
			break;

		orig_lines = orig_num_lines (p);
		new_lines = new_num_lines (line);

		p++;
		linenum = strtoul (p, &q, 10);
		if (p == q)
			break;
		if (linenum == 0)
			file_is_removed = 1;

		/* Now copy the relevant bits of the hunk */
		last_was_add = 0;
		newline = 1;
		while (orig_lines || new_lines || newline) {
			pos = ftell (f);
			if (getline (&line, &linelen, f) == -1)
				break;

			if (!orig_lines && !new_lines &&
			    line[0] != '\\')
				break;

			switch (line[0]) {
			case ' ':
				if (new_lines) new_lines--;
			case '-':
				if (orig_lines) orig_lines--;
				if (!file_is_removed)
					add_line (file, line + 1, linenum++);
				break;
			case '+':
				if (new_lines) new_lines--;
				break;
			case '\\':
				newline = 0;
				if (!file_is_removed && !last_was_add) {
					struct lines *at, *prev = NULL;
					char *p;
					for (at = file->head; at; at = at->next)
						prev = at;
					if (!prev)
						error (EXIT_FAILURE, 0,
						       "Garbled patch");
					p = prev->line;
					p += strlen (p) - 1;
					if (*p == '\n')
						*p = '\0';
				}
				break;
			}

			last_was_add = (line[0] == '+');
		}

		if (!newline) {
			pos = ftell (f);
			if (getline (&line, &linelen, f) == -1)
				break;
		}
	}

	if (line)
		free (line);

	return file->head;
}

static int output_reverted (FILE *p1)
{
	char *line;
	char *oldname;
	ssize_t linelen;
	long pos;

	oldname = NULL;
	linelen = 0;
	if (getline (&oldname, &linelen, p1) < 0)
		error (EXIT_FAILURE, errno, "Bad patch #1");

	if (strncmp (oldname, "--- ", 4))
		error (EXIT_FAILURE, 0, "Bad patch #1");

	line = NULL;
	linelen = 0;
	if (getline (&line, &linelen, p1) < 0)
		error (EXIT_FAILURE, errno, "Bad patch #1");

	if (strncmp (line, "+++ ", 4))
		error (EXIT_FAILURE, 0, "Bad patch #1");

	if (human_readable)
		printf ("reverted:\n");
	printf ("--- %s", line + 4);
	printf ("+++ %s", oldname + 4);
	free (oldname);

	pos = ftell (p1);
	if (getline (&line, &linelen, p1) < 0)
		error (EXIT_FAILURE, errno, "Bad patch #1");

	for (;;) {
		unsigned long orig_lines;
		unsigned long new_lines;
		unsigned long newline;
		char *d1, *d2, *p;
		size_t h;

		if (strncmp (line, "@@ ", 3)) {
			fseek (p1, pos, SEEK_SET);
			break;
		}

		p = line + 3;
		h = strcspn (p, " \t");
		d1 = xstrndup (p, h);
		p += h;
		p += strspn (p, " \t");
		h = strcspn (p, " \t");
		d2 = xstrndup (p, h);
		if (!*d1 || !*d2)
			error (EXIT_FAILURE, 0, "Bad patch #1");

		printf ("@@ -%s +%s @@\n", d2 + 1, d1 + 1);
		orig_lines = orig_num_lines (d1);
		new_lines = new_num_lines (d2);

		free (d1);
		free (d2);

		newline = 1;
		while (orig_lines || new_lines || newline) {
			pos = ftell (p1);
			if (getline (&line, &linelen, p1) == -1)
				break;

			if (!orig_lines && !new_lines &&
			    line[0] != '\\')
				break;

			switch (line[0]) {
			case ' ':
				if (orig_lines) orig_lines--;
				if (new_lines) new_lines--;
				printf ("%s", line);
				break;
			case '+':
				if (new_lines) new_lines--;
				printf ("-%s", line + 1);
				break;
			case '-':
				if (orig_lines) orig_lines--;
				printf ("+%s", line + 1);
				break;
			case '\\':
				newline = 0;
				printf ("%s", line);
				break;
			}
		}

		if (!newline) {
			pos = ftell (p1);
			if (getline (&line, &linelen, p1) == -1)
				break;
		}
	}

	if (line)
		free (line);

	return 0;
}

static int write_file (struct lines *file, int fd)
{
	unsigned long linenum;
	struct lines *at;
	FILE *fout = fdopen (fd, "w");

	linenum = 1;
	for (at = file; at; at = at->next) {
		unsigned long i = at->n - linenum;
		while (i--) {
			fputc ('\n', fout);
			linenum++;
		}
		fprintf (fout, "%s", at->line);
		linenum++;
	}
	fclose (fout);
	return 0;
}

static int apply_patch (FILE *patch, const char *file)
{
	unsigned long orig_lines, new_lines;
	ssize_t linelen;
	char *line;
	pid_t child;
	FILE *w;

	w = xpipe(PATCH, &child, "w", PATCH, "-sp0", NULL);

	fprintf (w, "--- %s\n+++ %s\n", file, file);
	line = NULL;
	linelen = 0;
	orig_lines = new_lines = 0;
	for (;;) {
		if (getline (&line, &linelen, patch) == -1)
			break;

		if (!orig_lines && !new_lines && !strncmp (line, "--- ", 4))
			break;

		fputs (line, w);

		if (!strncmp (line, "@@ ", 3)) {
			orig_lines = orig_num_lines (line);
			new_lines = new_num_lines (line);
			continue;
		}

		if (orig_lines && line[0] != '+')
			orig_lines--;
		if (new_lines && line[0] != '-')
			new_lines--;
	}
	fclose (w);
	waitpid (child, NULL, 0);

	if (line)
		free (line);

	return 0;
}

static int munge_diff_output (FILE *diff, const char *patched1,
			      const char *patched2)
{
	char *line = NULL;
	ssize_t linelen = 0;
	unsigned long context;
	const char *p;
	char *sub, *add;
	char *end;
	FILE *pf;

	if (getline (&line, &linelen, diff) == -1)
		return -1;

	sub = strchr (line, '-');
	add = strchr (line, '+');
	if (!sub || !add)
		error (EXIT_FAILURE, 0, "Garbled diff output");
	if (!strncmp (sub, "-0,0", 4))
		sub = NULL;
	if (!strncmp (add, "+0,0", 4))
		add = NULL;
	if (add && sub) {
		/* Normal case */
		printf ("%s", line);
		for (;;) {
			int ch = fgetc (diff);
			if (ch == EOF)
				break;
			putchar (ch);
		}
		return 0;
	}

	if (add)
		p = add + 1;
	else
		p = sub + 1;

	context = strtoul (p, &end, 10);
	if (end == p)
		error (EXIT_FAILURE, 0, "Garbled diff output");

	free (line);
	line = NULL;
	linelen = 0;

	if (add)
		/* Change @@ -0,0 +n @@ into @@ -1,1 +1,n+1 @@ */
		printf ("@@ -1,1 +1,%lu @@\n", context + 1);
	else
		/* Change @@ -n +0,0 @@ into @@ -1,n+1 +1,1 @@ */
		printf ("@@ -1,%lu +1,1 @@\n", context + 1);

	for (;;) {
		if (getline (&line, &linelen, diff) == -1) {
			free (line);
			line = NULL;
			linelen = 0;
			break;
		}
		if (line[0] == '@')
			break;
		printf ("%s", line);
	}

	/* Need one line of context */
	p = add ? patched2 : patched1;
	pf = xopen (p, "r");

	while (context) {
		int ch = fgetc (pf);
		if (ch == EOF)
			break;
		if (ch == '\n')
			context--;
	}

	if (context)
		error (EXIT_FAILURE, 0, "Need at least one line of context\n");

	putchar (' ');
	for (;;) {
		int ch = fgetc (pf);
		if (ch == EOF)
			break;
		putchar (ch);
		if (ch == '\n')
			break;
	}
	fclose (pf);

	if (line) {
		printf ("%s", line);
		free (line);
		for (;;) {
			int ch = fgetc (diff);
			if (ch == EOF)
				break;
			putchar (ch);
		}
	}

	return 0;
}

static int output_delta (FILE *p1, FILE *p2)
{
	const char *tmpdir = getenv ("TMPDIR");
	unsigned int tmplen;
	const char tail1[] = "/interdiff-1.XXXXXX";
	const char tail2[] = "/interdiff-2.XXXXXX";
	char *tmpp1, *tmpp2;
	int tmpp1fd, tmpp2fd;
	struct lines_info file = { NULL, NULL };
	struct lines_info file2 = { NULL, NULL };
	char *oldname = NULL, *newname = NULL;
	pid_t child;
	FILE *in;
	ssize_t namelen;
	long pos1 = ftell (p1), pos2 = ftell (p2);
	long start1, start2;
	char options[100];
	int patch_removes_file;
	int diff_is_empty = 1;

	if (!tmpdir)
		tmpdir = P_tmpdir;

	tmplen = strlen (tmpdir);
	tmpp1 = alloca (tmplen + sizeof (tail1));
	memcpy (tmpp1, tmpdir, tmplen);
	memcpy (tmpp1 + tmplen, tail1, sizeof (tail1));
	tmpp2 = alloca (tmplen + sizeof (tail2));
	memcpy (tmpp2, tmpdir, tmplen);
	memcpy (tmpp2 + tmplen, tail2, sizeof (tail2));

	if (max_context == 2)
		sprintf(options, "-%su",
			ignore_whitespace ? "w" : "");
	else
		sprintf (options, "-%sU%d",
			ignore_whitespace ? "w" : "", max_context);

	tmpp1fd = xmkstemp (tmpp1);
	tmpp2fd = xmkstemp (tmpp2);

	do {
		if (oldname) {
			free (oldname);
			oldname = NULL;
		}
		if (getline (&oldname, &namelen, p1) < 0)
			error (EXIT_FAILURE, errno, "Bad patch1");

	} while (strncmp (oldname, "+++ ", 4));
	oldname[strlen (oldname) - 1] = '\0';

	do {
		if (newname) {
			free (newname);
			newname = NULL;
		}
		if (getline (&newname, &namelen, p2) < 0)
			error (EXIT_FAILURE, errno, "Bad patch #2");

	} while (strncmp (newname, "+++ ", 4));
	newname[strlen (newname) - 1] = '\0';

	start1 = ftell (p1);
	start2 = ftell (p2);
	fseek (p1, pos1, SEEK_SET);
	fseek (p2, pos2, SEEK_SET);
	create_orig (p2, &file);
	fseek (p1, pos1, SEEK_SET);
	fseek (p2, pos2, SEEK_SET);
	create_orig (p1, &file2);
	merge_lines(&file, &file2);
	pos1 = ftell (p1);

	/* Write it out. */
	write_file (file.head, tmpp1fd);
	write_file (file.head, tmpp2fd);
	free_lines (file.head);

	fseek (p1, start1, SEEK_SET);
	fseek (p2, start2, SEEK_SET);

	apply_patch (p1, tmpp1);
	in = xopen (tmpp1, "r");
	patch_removes_file = (fgetc (in) == EOF);
	fclose (in);

	apply_patch (p2, tmpp2);
	if (!patch_removes_file) {
		in = xopen (tmpp2, "r");
		if (!in) {
			perror (tmpp2);
			exit (1);
		}
		patch_removes_file = (fgetc (in) == EOF);
		fclose (in);
	}
	fseek (p1, pos1, SEEK_SET);

	fflush (NULL);

	in = xpipe(DIFF, &child, "r", DIFF, options, tmpp1, tmpp2, NULL);
	
	/* Eat the first line */
	for (;;) {
		int ch = fgetc (in);
		if (ch == EOF || ch == '\n')
			break;
		diff_is_empty = 0;
	}

	/* Eat the second line */
	for (;;) {
		int ch = fgetc (in);
		if (ch == EOF || ch == '\n')
			break;
	}

	if (!diff_is_empty) {
		/* First character */
		char *p;
		p = strchr (oldname + 4, '\t');
		if (p) *p = '\0';
		p = strchr (newname + 4, '\t');
		if (p) *p = '\0';
		if (human_readable)
			printf (DIFF " %s %s %s\n", options,
				oldname + 4, newname + 4);
		printf ("--- %s\n", oldname + 4);
		printf ("+++ %s\n", newname + 4);
	}

	if (patch_removes_file) {
		for (;;) {
			int ch = fgetc (in);
			if (ch == EOF)
				break;
			putchar (ch);
		}
	} else munge_diff_output (in, tmpp1, tmpp2);

	fclose (in);
	waitpid (child, NULL, 0);
	unlink (tmpp1);
	unlink (tmpp2);
	free (oldname);
	free (newname);
	return 0;
}

static int copy_residue (FILE *p2)
{
	struct file_list *at;
	char *line = NULL;
	ssize_t linelen = 0;
	int first;

	for (at = files_in_patch2; at; at = at->next) {

		if (file_in_list (files_done, at->file) != -1)
			continue;

		/* check if we need to process it and init context */
		if (!check_filename(at->file))
			continue;

		fseek (p2, at->pos, SEEK_SET);
		first = 1;
		if (human_readable)
			printf ("only in patch2:\n");

		do {
			if (getline (&line, &linelen, p2) == -1)
				break;
			printf ("%s", line);
		} while (strncmp (line, "--- ", 4));
		if (getline (&line, &linelen, p2) == -1)
			break;
		printf ("%s", line);
		if (getline (&line, &linelen, p2) == -1)
			break;
		for (;;) {
			unsigned long orig_lines, new_lines, newline;
			if (strncmp (line, "@@ ", 3))
				break;
			printf ("%s", line);
			orig_lines = orig_num_lines (line);
			new_lines = new_num_lines (line);
			newline = 1;
			while (orig_lines || new_lines || newline) {
				if (getline (&line, &linelen, p2) == -1)
					break;
				
				if (!orig_lines && !new_lines &&
				    line[0] != '\\')
					break;

				switch (line[0]) {
				case ' ':
					if (orig_lines) orig_lines--;
					if (new_lines) new_lines--;
					break;
				case '-':
					if (orig_lines) orig_lines--;
					break;
				case '+':
					if (new_lines) new_lines--;
					break;
				case '\\':
					newline = 0;
					break;
				}
				printf ("%s", line);
			}

			if (!newline &&
			    getline (&line, &linelen, p2) == -1)
				break;
		}
	}

	free (line);
	return 0;
}

static int index_patch2 (FILE *p2)
{
	char *line = NULL;
	ssize_t linelen = 0;

	/* Index patch2 */
	while (!feof (p2)) {
		unsigned long skip;
		char *names[2];
		char *p, *end;
		size_t h;
		long pos = ftell (p2);

		if (getline (&line, &linelen, p2) == -1)
			break;

		if (strncmp (line, "--- ", 4))
			continue;

		h = strcspn (line + 4, " \t");
		names[0] = xstrndup (line + 4, h);

		if (getline (&line, &linelen, p2) == -1) {
			free (names[0]);
			break;
		}

		if (strncmp (line, "+++ ", 4)) {
			free (names[0]);
			continue;
		}

		h = strcspn (line + 4, " \t");
		names[1] = xstrndup (line + 4, h);

		if (getline (&line, &linelen, p2) == -1) {
			free (names[0]);
			free (names[1]);
			break;
		}

		if (strncmp (line, "@@ ", 3))
			goto try_next;

		p = strchr (line + 3, '+');
		if (!p)
			goto try_next;
		p = strchr (p, ',');
		if (p) {
			/* Like '@@ -1,3 +1,3 @@' */
			p++;
			skip = strtoul (p, &end, 10);
			if (p == end)
				goto try_next;
		} else
			/* Like '@@ -1 +1 @@' */
			skip = 1;

		add_to_list (&files_in_patch2, best_name (2, names), pos);

		while (skip--) {
			if (getline (&line, &linelen, p2) == -1) {
				fprintf (stderr, "Hmm, patch2 looks odd\n");
				break;
			}
			if (line[0] == '-')
				/* Doesn't count towards skip count */
				skip++;
		}

	try_next:
		free (names[0]);
		free (names[1]);
	}

	if (line)
		free (line);

	return 0;
}

static int interdiff (FILE *p1, FILE *p2)
{
	char *line = NULL;
	ssize_t linelen = 0;

	index_patch2 (p2);

	/* Search for next file to patch */
	while (!feof (p1)) {
		char *names[2];
		char *p;
		size_t h;
		long pos, start_pos = ftell (p1);

		if (line) {
			free (line);
			line = NULL;
			linelen = 0;
		}

		if (getline (&line, &linelen, p1) == -1)
			break;

		if (strncmp (line, "--- ", 4))
			continue;

		h = strcspn (line + 4, " \t");
		p = xstrndup (line + 4, h);
		names[0] = p;

		if (getline (&line, &linelen, p1) == -1)
			break;

		if (strncmp (line, "+++ ", 4))
			continue;

		h = strcspn (line + 4, " \t");
		p = xstrndup (line + 4, h);
		names[1] = p;

		p = xstrdup (best_name (2, names));
		free (names[0]);
		free (names[1]);
		
		/* check if we need to process it and init context */
		if (!check_filename(p)) {
			add_to_list (&files_done, p, 0);
			continue;
		}

		fseek (p1, start_pos, SEEK_SET);
		pos = file_in_list (files_in_patch2, p);
		if (pos == -1) {
			output_reverted (p1);
		} else {
			fseek (p2, pos, SEEK_SET);
			output_delta (p1, p2);
		}

		add_to_list (&files_done, p, 0);
	}

	copy_residue (p2);

	free_list (files_in_patch2);
	free_list (files_done);
	if (line)
		free (line);
	return 0;
}

NORETURN
static void syntax (int err)
{
	const char *const syntax_str =
"usage: %s [OPTIONS] patch1 patch2\n"
"       %s --version|--help\n"
"OPTIONS are:\n"
"  -U N            max lines of context to carry\n"
"  -w              ignore whitespace changes in patches\n"
"  -p N            pathname components to ignore\n"
"  -d PAT          drop context on matching files\n"
"  -z              decompress .gz and .bz2 files\n";

	fprintf (err ? stderr : stdout, syntax_str, progname, progname);
	exit (err);
}

int main (int argc, char *argv[])
{
	FILE *p1, *p2;
	int ret;

	set_progname (argv[0]);
	for (;;) {
		static struct option long_options[] = {
			{"help", 0, 0, 'H'},
			{"version", 0, 0, 'v'},
			{0, 0, 0, 0}
		};
		char *end;
		int c = getopt_long (argc, argv, "U:hp:i:I:x:X:d:zw",
				     long_options, NULL);
		if (c == -1)
			break;

		switch (c) {
		case 'v':
			printf("interdiff - patchutils version %s\n", VERSION);
			exit(0);
		case 'H':
			syntax (0);
			break;
		case 'h':
			break;
		case 'U':
			max_context_real = strtoul (optarg, &end, 0);
			if (optarg == end)
				syntax (1);
			break;
		case 'p':
			ignore_components = strtoul (optarg, &end, 0);
			if (optarg == end)
				syntax (1);
			break;
		case 'd':
			patlist_add (&pat_drop_context, optarg);
			break;
		case 'z':
			unzip = 1;
			break;
		case 'w':
			ignore_whitespace = 1;
			break;
		default:
			syntax(1);
		}
	}
	
	if (optind + 2 != argc)
		syntax (1);
	
	if (unzip) {
		p1 = xopen_unzip (argv[optind], "rb");
		p2 = xopen_unzip (argv[optind + 1], "rb");
	} else {
		p1 = xopen_seekable (argv[optind], "rb");
		p2 = xopen_seekable (argv[optind + 1], "rb");
	}

	ret = interdiff (p1, p2);

	fclose (p1);
	fclose (p2);
	patlist_free (&pat_drop_context);
	return 0;
}
